/*
 * Copyright (c) 1997 Manuel Bouyer.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Manuel Bouyer.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
 * poppassd.c  Update passwords on the UNIX system from the Eudora
 *             "Change Password"
 *             client.  Keeps a log of change attempts in /var/adm/poppassd.log
 *             but suppresses recording of passwords given by the client.
 *             
 *             Must be owned by root, and executable only by root.  It can be
 *             started with an entry in /etc/inetd.conf such as the following:
 *
 * poppassd stream tcp nowait root  /usr/local/poppassd/poppassd poppassd
 * 
 *             and in /etc/services:
 * 
 * ppassd        106/tcp
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <pwd.h>
#include <signal.h>
#include <unistd.h>
#include <varargs.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/file.h>
#include <time.h>
#include "lib.h"

#ifdef __NetBSD__
#define LOGFILE "/var/log/poppasswd.log"
#else
#define LOGFILE "/var/adm/poppasswd.log"
#endif

main (argc, argv)
int argc;
char *argv[];
{

		static char line[BUFSIZE];
		static char user[BUFSIZE];
		static char pass[BUFSIZE];
		static char newpass[BUFSIZE];
		static char logFile[BUFSIZE];
        struct passwd *pw;

        chdir( "/etc" );
		sprintf(logFile, LOGFILE);
        strcpy( user, "" );
        strcpy( pass, "" );
        strcpy( newpass, "" );

        /*
         * Open the logfile for append
         */
        if ( !( log = fopen( logFile, "a+" )) ) {
                WriteToClient( "500 Can't open %s.", logFile );
                exit(1);
        }
        
		/*
         * The server's responses should be like an FTP server's responses; 
         * 1xx for in progress, 2xx for success, 3xx for more information
         * needed, 4xx for temporary failure, and 5xx for permanent failure.  
         * Putting it all together, here's a sample conversation:

         * S: 200 hello\r\n
         * E: user yourloginname\r\n
         * S: 300 please send your password now\r\n
         * E: pass yourcurrentpassword\r\n
         * S: 200 My, that was tasty\r\n
         * E: newpass yournewpassword\r\n
         * S: 200 Happy to oblige\r\n
         * E: quit\r\n
         * S: 200 Bye-bye\r\n
         * S: <closes connection>
         * E: <closes connection>
		 */

        WhoAreYou();

        WriteToClient( "200 hello, who are you?" );
        ReadFromClient( line, 0 );
        sscanf(line, "user %s", user) ;
        if ( strlen(user) == 0 ) {
                WriteToClient( "500 username required." );
                exit(1);
        }

        WriteToClient( "200 your password please." );
        ReadFromClient( line, SUPPRESS );
        sscanf( line, "pass %s", pass) ;
        if ( strlen(pass) == 0 ) {
                WriteToClient( "500 password required." );
                exit(1);
        }

        WriteToClient( "200 your new password please." );
        ReadFromClient( line, SUPPRESS );
        sscanf( line, "newpass %s", newpass );

        /* new pass required */
        if ( strlen(newpass) == 0 ) {
                WriteToClient( "500 new password required." );
                exit(1);
        }
        /* new pass must be 6 char or longer */
        else if ( strlen(newpass) < 6 ) {
                WriteToClient( "500 New password too short" );
                exit(1);
        }
        /* new pass must be different from old pass */
        else if ( !strcmp(pass, newpass) ) {
                WriteToClient( "500 New password must be different." );
                exit(1);
        }

        /* test for valid user */
        if ( (pw=getpoppwnam(user)) == NULL ) {
                WriteToClient( "500 Unknown user, %s.", user);
                exit(1);
        }

        /* authenticate user */
        if ( chkPass( user, pass, pw ) == FAILURE ) {
                WriteToClient( "500 Authentication failure.");
                exit(1);
        }
                
        if ( setPass( user, pass, newpass, pw ) == FAILURE ) { 
                WriteToClient( "500 Unable to change password." );
                exit(1);
        }

        WriteToClient( "200 Password changed, thank-you." );
        return(1);
}

 
WhoAreYou()
{
        struct sockaddr_in bobby;
        int   bobbyLen = sizeof(struct sockaddr);
        struct passwd *pw;
        char   *inet_ntoa();
        char   *ttyname();

        fputs( "\n", log );

        if (isatty(fileno(stdin))) {
                pw = getpwuid(getuid());
                WriteLog( "Connection on %s by %s",
                        ttyname(fileno(stdin)), pw->pw_name);
        }

        else if (lseek(fileno(stdin), 0, SEEK_CUR) >= 0) {
                pw = getpwuid(getuid());
                WriteLog( "Connection on FILE by %s", pw->pw_name);
        }

        else if (getpeername(fileno(stdin), (struct sockaddr *)&bobby, 
                        &bobbyLen) < 0) {
                pw = getpwuid(getuid());
                WriteLog( "Connection on PIPE by %s", pw->pw_name);
        }

        else {
#               if defined(sparc) && __GNUC__ == 1 
                        WriteLog( "NET connection from  %s", 
                                inet_ntoa(&bobby.sin_addr));
#               else
                        WriteLog( "NET connection from %s", 
                                inet_ntoa(bobby.sin_addr));
#               endif
        }
}

WriteToClient( va_alist )
va_dcl
{
        va_list ap;
        char *fmt;      
        char *args[MAXARGS];
        int argno = 0;
static  char string[BUFSIZE];
        
        UpdateTime( );

        va_start( ap );
        fmt = va_arg(ap, char *);
        while ((args[argno++] = va_arg(ap, char *)) != (char *)0)
                    ;

        vfprintf(stdout, fmt, args);
        fputs("\r\n", stdout );
        fflush(stdout);

        vsprintf(string, fmt, args);
        WriteLog( "Server> %s", string );

        va_end( ap );
}

ReadFromClient( line , suppress)
char *line;
int suppress;
{
        char *sp;
        int i;

        UpdateTime();
        strcpy(line, "" );
        fgets( line, BUFSIZE, stdin );
        if (( sp = strchr(line, '\n')) != NULL ) *sp = '\0'; 
        if (( sp = strchr(line, '\r')) != NULL ) *sp = '\0'; 
        if ( suppress )
                WriteLog( "Client> ..." );
        else
                WriteLog( "Client> %s", line );
}
